/* jasenet.cpp */
/****************************************************************************/
/*
**        J A S E N E T . C P P
**
** Jsenet-luokan toteutus
**
**  Tekijt:          Vesa Lappalainen
**                    Ohjelmointikurssi 1991
**  Tehty:            20.12.1991
**  Muutettu:         09.12.1995
**  Mit muutettu:    Muutettu C++ luokaksi
**  Muutettu:         10.12.1995
**  Mit muutettu:    + Laitettu tiedoston lukeminen ja kirjoittaminen
**                      toimivaksi (ks. metodit lue_tiedostosta ja talleta)
**                    + Jsenen lisys kasvattaa taulukon kokoa, mikli
**                      tila loppuu (ks. lisaa)
**  Muutettu:         20.01.1997
**  Mit muutettu:    + Otettu kerhosta jsenistn osuus
**
** Public-metodeja:
** ================
**
**  Metodit palauttavat yleens virheen merkkijonona, joka voidaan
**  suoraan tulostaa.  Jos virhett ei tapahdu, palautetaan NULL.
**
**  cJasenet(int koko=0)
**    - jsenitn luominen, jos koko = 0, ei luoda jsentaulukkoa
**
**  ~cJasenet()
**    - jsenistn hvittminen, jos jsenist ei ole talletettu, talletetaan
**      viimeistn nyt (mahd. virheilmoitus harakoille)
**
**  string lisaa(const cJasen &jasen);
**    - listn uusi jsen.  Jsenest tehdn kopio.  Jos tila loppuu,
**      yritetn kasvattaa tilaa.
**
**  string korvaa(int i,const cJasen &jasen);
**    - korvataan i:s jsen uudella.  Jos i vrin, ei tehd mitn (pal 1).
**
**  const cJasen &anna(int i) const
**    - otetaan kerhosta i:s jsen.  Jos i vrin, otetaan 0:s jsen.
**      Kerholaisia voidaan muuttaa, kun otetaan jsenest kopio
**      anna-metodilla, muutetaan kopion tietoja ja laitetaan takaisin
**      korvaa-metodilla:
**        cJasen jasen(kerho.anna(i));
**        - muuta jasenta
**        kerho.korvaa(i,jasen);
**
**  string luo(const string &tied,const string &nimi,
**    - luo jsenistn uutta kerhoa varten, mahd. vanha talletetaan
**
**  string lue(const string &tied);
**    - luetaan uusi jsenist, mahd. vanha talletetaan
**
**  string talleta(const string &tied="");
**    - talletetaan jsensist.  Jos tied=="", kytetn luettaessa
**      annettua nimi.
**
**  const string &Koko_nimi() const
**    - palauttaa tiedoston koko nimen.
**
**  int Lkm() const
**    - palauttaa jsenmrn
**
**  const string &Tiedoston_nimi() const
**    - paluttaa viimeksi kytetyn tiedoston nimen
**
**  int Muutettu() const
**    - palauttaa 1, mikli jsenist on muutettu talletusten jlkeen ja
**      0 muuten.
**
**  int cJasenet::TeeBak(const string &bak_tark)
**    - palauttaa 0, mikli tiedoston muuttaminen bak-tiedostoksi onnistuu
**      1 muuten.
**
**  int laske_montako_muuta(const cJasen &jasen, int k,int &kuka)
**    - laskee monellako jsenell on kentss k samat tiedot
**      kuin jasenella.  Palautetaan mys viimeinen ko. jsen.
** ---------------------------------------------------------------------------
**
** Tietorakenne on seuraava:
** =========================
**
**
**   cJasenet                       cJasen
**   ------
**   |    |                         ------
**   |    |                         |    |
**   | 7  | max               |---->|    |
**   | 3  |       alkiot      |     |    |  ------
**   | o--+------>|-----|     |     ------  |    |
**   ------       |  o--+-----|             |    |
**                |-----|            |----->|    |
**                |  o--+------------|      ------   ------
**                |-----|                            |    |
**                |  o--+--------------------------->|    |
**                |-----|                            |    |
**                |  o--+--?                         ------
**                |-----|
**                |  o--+--?
**                |-----|
**                |  o--+--?
**                |-----|
**                |  o--+--?
**                |-----|
**
**
** Jsenet ei tied jseniens ominaisuuksia.
** Jsenist pit huolta siit, onko jsenet muuttunut, eli tarvitseeko se
** talletusta vaiko ei.
** Tiedoston muoto, ks. lue_tiedostosta - metodi
**
*/
#include <cstdlib>
#include <fstream>
#include "mjonotpp.h"
#include "jasenet.h"
using std::ifstream;
using std::ofstream;

//============================================================================
// Makrot:
//============================================================================


//----------------------------------------------------------------------------
// Makro, jolla tutkitaan palauttiko kutsu v virheen vai ei ja jos
// palautti, niin poistutaan aliohjelmasta palauttaen sama virhe.
#define IF_ERR_RETURN(v) { string virhe=(v); if ( virhe != "" ) return virhe; }

//============================================================================
// Virheilmoitukset:
//============================================================================
static const char *EI_VOI_LUODA = "Taulukolle ei saada varattua tilaa muistista!";
static const char *LIIKAA_ALKIOITA = "Liikaa alkioita, ei voi list!";
static const char *EI_SAA_ALKIOTA = "Ei saada tilaa uudelle alkiolle muistista!";
static const char *EI_NIMEA       = "Ei koko nime tiedostossa!";
static const char *EI_MAXKOKOA    = "Ei maksimikokoa tiedoston alussa!";
static const char *TIED_EI_AUKEA  = "Tiedosto ei aukea!";
static const char *OTS_EI_KIRJ    = "Otsikkotietoja ei voi kirjoittaa!";
static const char *ALKIO_EI_KIRJ  = "Alkiota ei voi kirjoittaa!";
static const char *LAITON_INDEKSI = "Laiton indeksi!";
static const char *EI_REKISTEROITY= "Alkiota ei ole rekisterity!";

//============================================================================
// Private -metodeja:
//============================================================================

//----------------------------------------------------------------------------
void cJasenet::poista_taulukko()
{
  if ( max_lkm > 0 ) free(alkiot);
  max_lkm = 0;
}


//----------------------------------------------------------------------------
void cJasenet::poista_alkiot()
{
  for (int i=0; i<lkm; i++)
    delete alkiot[i];
  lkm = 0;
}


//----------------------------------------------------------------------------
string cJasenet::kasvata_kokoa()
/*
** Yritetn allokoida uutta tilaa 50% maksimimrn nhden lis.
** Paitsi jos vanha koko on 1, niin kasvatetaan 2:ksi.
** Jos vanha tila on 0, niin tehdn uusi tila.
----------------------------------------------------------------------------*/
{
  if ( max_lkm <= 0 ) return luo_taulukko(2);
  int uusi_koko = 3*(max_lkm)/2; if ( uusi_koko <= 1 ) uusi_koko=2;

  cJasen **uusi_tila = (cJasen **)realloc(alkiot,uusi_koko*sizeof(cJasen *));

  if ( uusi_tila == NULL ) return LIIKAA_ALKIOITA;
  alkiot = uusi_tila;
  max_lkm = uusi_koko;
  return "";
}


//============================================================================
// Jsenet-luokan public-metodien toteutus:
//============================================================================

//----------------------------------------------------------------------------
string cJasenet::lisaa(const cJasen &jasen)
/*
** Listn uusi jsen kerhoon.  EI tarkisteta esiintyyk
** nimi jo ennestn.  Mikli tilaa ei ole, yritetn allokoida uutta
** tilaa 50% maksimimrn nhden lis.
**
** Algoritmi: Jos uudelle jsenelle saadaan varattua tilaa, listn
**            uuden paikan osoite taulukkoon ja kopioidaan tiedot
**            uuteen paikkaan.
----------------------------------------------------------------------------*/
{
  if ( lkm >= max_lkm ) IF_ERR_RETURN(kasvata_kokoa());

  try {
    cJasen *uusi = new cJasen(jasen);
    alkiot[lkm] = uusi;
    lkm++;

    muutettu = 1;
  }
  catch ( std::bad_alloc ) {
    return EI_SAA_ALKIOTA;
  }

  return "";
}

//----------------------------------------------------------------------------
string cJasenet::korvaa(int i,const cJasen &uusi)
/*
** Korvataan paikassa i oleva jsen uudella jsenell.
** Vain rekisterityneet jsenet hyvksytn.
** Jos samat tiedot kuin vanhalla jsenell, ei korvata.
----------------------------------------------------------------------------*/
{
  if ( ulkona(i) ) return LAITON_INDEKSI;
  if ( alkiot[i]->sama_rekisteri(uusi) == 0 ) return EI_REKISTEROITY;
  if ( *(alkiot[i]) == uusi ) return "";
  muutettu = 1;
  *(alkiot[i]) = uusi;
  return "";

}


//----------------------------------------------------------------------------
string cJasenet::luo_taulukko(int koko)
/*
** Luodaan jsentaulukko valitun kokoiseksi.  Jos jsentaulukko on ennestn
** on siell on se parasta poistaa.
** Luomisessa kytetn malloc-funktiota, jotta koon muuttaminen on
** jatkossa helpompaa realloc-funktiolla. (ks. lisaa-metodi).
----------------------------------------------------------------------------*/
{
  poista_kaikki();
  if ( koko <= 0 ) return "";
  alkiot = (cJasen **)malloc( koko * (sizeof(cJasen *)) );
  lkm = 0;
  max_lkm = 0;
  if ( alkiot == NULL ) return EI_VOI_LUODA;
  max_lkm = koko;
  return "";
}


//----------------------------------------------------------------------------
string cJasenet::luo(const string &tied,const string &nimi,int max_koko)
/*
** Luodaan uusi jsenist valitulla tiedoston nimell ja jsenistn nimell.
** Taulukon ylrajaksi luodaan max_jas.
----------------------------------------------------------------------------*/
{
  IF_ERR_RETURN(luo_taulukko(max_koko));

  tiedoston_nimi = tied;
  koko_nimi      = nimi;

  muutettu = 1;
  return "";
}


//----------------------------------------------------------------------------
string cJasenet::lue_tiedostosta(const string &tied)
/*
** Luetaan tiedostosta jsenistn tiedot.
**
** Metodit olettavat tiedoston olevan seuraavaa muotoa:
**
**Kelmien kerho ry
**50
**; Kenttien jrjestys tiedostossa on seuraava:
**id|sukunimi etunimi  |sotu       |katuosoite  |postinumero|postiosoite|koti...
**1|Ankka Aku          |010245-123U|Ankkakuja 6 |12345      |ANKKALINNA |12-1...
**2|Susi Sepe          |020347-123T|            |12555      |Permets  |    ...
**3|Ponteva Veli       |030455-3333|            |12555      |Permets  |    ...
**<EOF>
**
** Tiedoston 3. rivilt alkaen olevat kommenttirivit (merkitty ;) unohdetaan.
**
** Taulukon ylrajaksi luodaan tiedosta 2. rivilt luettu max_koko.
----------------------------------------------------------------------------*/
{
  ifstream f(tied.c_str());
  if ( !f ) return TIED_EI_AUKEA;

  string nimi;  lue_rivi(f,nimi);       if ( !f ) return EI_NIMEA;
  int max_koko; lue_rivi(f,max_koko);   if ( !f ) return EI_MAXKOKOA;

  IF_ERR_RETURN(luo_taulukko(max_koko));

  tiedoston_nimi = tied;
  koko_nimi      = nimi;

  string rivi;
  cJasen uusi;
  while ( std::getline(f,rivi) ) {
    if ( rivi == "" || rivi[0] == ';' ) continue;
    uusi.setAsString(rivi);
//    f >> uusi;  // vaatisi kunkin rivin olemisen tydellisen, muuten OK!
    IF_ERR_RETURN(lisaa(uusi));
  }

  muutettu = 0;
  return "";
}


//----------------------------------------------------------------------------
string cJasenet::talleta(const string &tied)
{
  if ( !muutettu ) return "";
  string tiedosto(tied); if ( tied == "" ) tiedosto = tiedoston_nimi;
  ofstream f(tiedosto.c_str());

  if ( !f ) return TIED_EI_AUKEA;

  f << koko_nimi << "\n";
  f << max_lkm << "\n";

  if ( !f ) return OTS_EI_KIRJ;

  for (int i=0; i<lkm; i++) {
    f << *alkiot[i] << "\n";
    if ( !f ) return ALKIO_EI_KIRJ;
  }

  muutettu = 0;
  tiedoston_nimi = tiedosto;
  return "";
}

//----------------------------------------------------------------------------
int cJasenet::TeeBak(const string &bak_tark)
{
  bak_nimi = Tiedoston_nimi();
  vaihda_tarkennin(bak_nimi,bak_tark);

  remove(bak_nimi); /* Vanha .BAK tytyy poistaa jotta rename toimii */
  return rename(Tiedoston_nimi(),bak_nimi);
}

//----------------------------------------------------------------------------
int cJasenet::laske_montako_muuta(const cJasen &jasen, int k,int &kuka) const
// Lasketaan monnellako muulla kerholaisella on sama tieto kentss k
// kuin jasenella.  Palautetaan muiden maara ja sijoitetaan muuttujaan
// kuka viimeinen sellainen jolla oli sama tieto
{
  int i,samoja=0;
  string kentta = jasen.kentta_jonoksi(k);
  for (i=0; i<lkm; i++) {
    if ( alkiot[i]->kentta_jonoksi(k) == kentta &&
         alkiot[i]->sama_rekisteri(jasen) == 0 ) {
      samoja++;
      kuka = i;
    }
  }
  return samoja;
}

//----------------------------------------------------------------------------
void cJasenet::poista_ind(int ind)
{
  if ( ulkona(ind) ) return;
  muutettu = 1;
  delete alkiot[ind];
  for (int i=ind; i<lkm-1; i++)
    alkiot[i] = alkiot[i+1];
  lkm--;  
}


class cVertaa {
  const cJasenet *jasenet;
  int kentta;
 public:
  cVertaa(const cJasenet *ijasenet, int k) : jasenet(ijasenet), kentta(k) { }
  bool operator() (int a,int b) const { // a < b
//    string sa = jasenet->anna(a).kentta_avaimeksi(kentta);
    return jasenet->anna(a).kentta_jonoksi(kentta) <
           jasenet->anna(b).kentta_jonoksi(kentta);
  }
};

#include <algorithm>
//----------------------------------------------------------------------------
void cJasenet::etsi(cPermutaatio &loytyneet,int k,const string &jono)
{
  for (int i=0; i<lkm; i++)
    if ( onko_samat(anna(i).kentta_jonoksi(k),jono) == 0  )
      loytyneet.push_back(i);
  std::stable_sort(loytyneet.begin(),loytyneet.end(),cVertaa(this,k));
}


